/*

Copyright 1998 by Thorsten Knabe <tek@rbg.informatik.tu-darmstadt.de>

Based on the CS4232/AD1848 driver Copyright (C) by Hannu Savolainen 1993-1996
 
USS/Lite for Linux is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
Version 2 (June 1991). See the "COPYING" file distributed with this software
for more info.

version: 0.3 (first public release)
status: alpha
date: 1998/05/23
*/


#define DEBUGINFO(x) 
#define DEBUGLOG(x) x
#define DEBUGWARN(x) x

#define CHECK_FOR_POWER { int timeout=100; \
  while (timeout > 0 && (inb(devc->base)&0x80)!= 0x80) \
  timeout--; \
  if (timeout==0) \
    printk("Check for power failed in %s line: %d\n",__FILE__,__LINE__); \
}



#include <linux/config.h>
#include "sound_config.h"

#ifdef CONFIG_AD1816

/* structure to hold device specific information */
typedef struct
  {
    int             base;
    int             irq;
    int             dual_dma;	/* 1, when two DMA channels allocated */
    int             debug_flag;

    int             speed;
    int             channels;
    int             audio_format;
    unsigned char   format_bits;

    int             xfer_count;
    int             audio_mode;
    int             intr_active;
    int             opened;

    int             recmask;
    int             supported_devices;
    int             supported_rec_devices;
    unsigned short  levels[64];
    int             dev_no;
    volatile unsigned long timer_ticks;
    int             irq_ok;
    int             *osp;

    /* Mixer parameters */
    unsigned char   saved_regs[64];
  }

ad1816_info;

static int      nr_ad1816_devs = 0;

/* for backward mapping of irq to sound device */

static volatile char irq2dev[17] =
{-1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1};


/* supported audio formats */
static int      ad_format_mask =
  AFMT_U8 | AFMT_S16_LE | AFMT_S16_BE | AFMT_MU_LAW | AFMT_A_LAW;

/* array of device info structures */
static ad1816_info dev_info[MAX_AUDIO_DEV];


/* ------------------------------------------------------------------- */

/* functions for easier access to inderect registers */

static int ad_read (ad1816_info * devc, int reg)
{
  unsigned long   flags;
  int result;

  CHECK_FOR_POWER;

  save_flags (flags); /* make register access atomic */
  cli ();
  outb ((unsigned char) (reg & 0x3f), devc->base+0);
  result = inb(devc->base+2);
  result+= inb(devc->base+3)<<8;
  restore_flags (flags);

  return (result);
}


static void ad_write (ad1816_info * devc, int reg, int data)
{
  unsigned long flags;

  CHECK_FOR_POWER;

  save_flags (flags); /* make register access atomic */
  cli ();
  outb ((unsigned char) (reg & 0xff), devc->base+0);
  outb ((unsigned char) (data & 0xff),devc->base+2);
  outb ((unsigned char) ((data>>8)&0xff),devc->base+3);
  restore_flags (flags);

}

/* ------------------------------------------------------------------- */

/* function interface required by struct audio_driver */

static void ad1816_output_block (int dev, unsigned long buf, 
				 int count, int intrflag, int dma_restart)
{
  unsigned long flags;
  unsigned long cnt;
  unsigned char buffer;
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;

  DEBUGINFO(printk("ad1816_output_block called buf=%d count=%d flags=%d restart=%d\n",buf,count,intrflag,dma_restart));
  
  if (!dma_restart)
    return;

  cnt = (count/4)-1; /* dma transfer count */

  save_flags (flags); /* make register access atomic */
  cli ();
  if (devc->audio_mode & PCM_ENABLE_OUTPUT && audio_devs[dev]->flags & DMA_AUTOMODE &&
      intrflag &&
      cnt == devc->xfer_count)
    {
      devc->audio_mode |= PCM_ENABLE_OUTPUT;
      devc->intr_active = 1;
      restore_flags (flags);
      DEBUGWARN(printk("Automode not supported by this device\n"));
      return;	
    }

  buffer=inb(devc->base+8);
  if (buffer & 0x01) /* Playback enabled? */
    outb(buffer & ~0x01,devc->base+8);	/* Disable playback*/

  if (dma_restart) /* start dma channel */
    {
      DMAbuf_start_dma (dev, buf, count, DMA_MODE_WRITE);
    }
  ad_write (devc, 8, cnt & 0xffff); /* set transfer count */

  if (buffer & 0x01) /* was playback enabled? */
    outb(buffer, devc->base+8);	/* reset playback register to old value */
  
  devc->xfer_count = cnt;
  devc->audio_mode |= PCM_ENABLE_OUTPUT; /* now we are ready to play data */
  devc->intr_active = 1;
  restore_flags (flags);
}


static void ad1816_start_input (int dev, unsigned long buf, int count,
				int intrflag, int dma_restart)
{
  unsigned long flags;
  unsigned long  cnt, buffer;
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;

  DEBUGLOG(printk("ad1816_output_block called buf=%d count=%d flags=%d restart=%d\n",buf,count,intrflag,dma_restart));

  if (!dma_restart)
    return;

  cnt = (count/4)-1; /* dma transfer count */

  save_flags (flags); /* make register access atomic */
  cli ();
  if (devc->audio_mode & PCM_ENABLE_INPUT && audio_devs[dev]->flags & DMA_AUTOMODE &&
      intrflag &&
      cnt == devc->xfer_count)
    {
      devc->audio_mode |= PCM_ENABLE_INPUT;
      devc->intr_active = 1;
      restore_flags (flags);
      DEBUGWARN(printk("Automode not supported by this device\n"));
      return;			/*
				 * Auto DMA mode on. No need to react
				 */
    }



  buffer=inb(devc->base+9);	
  if (buffer & 0x01) /* Capture enabled? */
    outb(buffer & ~0x01,devc->base+9);

  if (dma_restart)
    {
      DMAbuf_start_dma (dev, buf, count, DMA_MODE_READ);
    }

  ad_write (devc, 10, cnt & 0xffff); /* set transfer count */

  if (buffer & 0x01) /* was capture enabled? */
    outb(buffer, devc->base+9);	/* reset capture register to old value */

  devc->xfer_count = cnt;
  devc->audio_mode |= PCM_ENABLE_INPUT;
  devc->intr_active = 1;
  restore_flags (flags);
}


static int ad1816_ioctl (int dev, unsigned int cmd, caddr_t arg, int local)
{
  return -(EINVAL);
}


static int ad1816_prepare_for_input (int dev, int bsize, int bcount)
{
  unsigned long flags;
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;
  unsigned char fmt_bits;

  DEBUGLOG (printk ("Prepare for input called: bsize=%d bcount=%d\n",bsize,bcount));

  save_flags (flags); /* make register access atomic */
  cli ();

  fmt_bits= (devc->format_bits&0x7)<<3;

  if (devc->channels > 1) /* set mono/stereo mode */
    {
      fmt_bits |=0x4;
    }

  outb( (inb(devc->base+8) & ~0x3C)|fmt_bits, devc->base+8); /* Mono */
  outb( (inb(devc->base+9) & ~0x3C)|fmt_bits, devc->base+9);
  
  ad_write (devc, 2, devc->speed & 0xffff);	/* set speeds */
  ad_write (devc, 3, devc->speed & 0xffff);	/* set speeds */

  devc->xfer_count = 0;
  restore_flags (flags);

  return 0;
}

static int ad1816_prepare_for_output (int dev, int bsize, int bcount)
{
  return ad1816_prepare_for_input (dev, bsize, bcount);
}

static void ad1816_trigger (int dev, int state) 
{
  unsigned long flags;
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;

  DEBUGLOG (printk("ad1848_trigger called!\n"));

  /* mode may have changed */

  save_flags (flags); /* make register access atomic */
  cli ();
  state &= devc->audio_mode; /* mask out those states we are not prepared for
				set soundchip to new io-mode */
  if (state & PCM_ENABLE_INPUT)
    outb(inb(devc->base+9)|0x01, devc->base+9);
  else
    outb(inb(devc->base+9)&~0x01, devc->base+9);
    
  if (state & PCM_ENABLE_OUTPUT)
    outb(inb(devc->base+8)|0x01, devc->base+8);
  else
    outb(inb(devc->base+8)&~0x01, devc->base+8);
  restore_flags (flags);
}

static void ad1816_halt_input (int dev)
{
  unsigned long flags;
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;
  unsigned char buffer;

  DEBUGLOG (printk("ad1816_halt_input called\n"));

  save_flags (flags); /* make register access atomic */
  cli ();
  buffer=inb(devc->base+9);
  if (buffer & 0x01)
    outb(buffer & ~0x01,devc->base+9); /* disable capture */

  outb (~0x40, devc->base+1);	/* Clear interrupt status */

  devc->audio_mode &= ~PCM_ENABLE_INPUT;
  restore_flags (flags);
}

static void
ad1816_halt_output (int dev)
{
  unsigned long flags;
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;
 
  unsigned char buffer;

  DEBUGLOG (printk("ad1848_halt_output called!\n"));

  save_flags (flags); /* make register access atomic */
  cli ();
  buffer=inb(devc->base+8);
  if (buffer & 0x01)
    outb(buffer & ~0x01,devc->base+8); /* disable capture */

  outb ((unsigned char)~0x80, devc->base+1);	/* Clear interrupt status */

  devc->audio_mode &= ~PCM_ENABLE_OUTPUT;
  restore_flags (flags);
}

static void ad1816_halt (int dev)
{
  ad1816_halt_input(dev);
  ad1816_halt_output(dev);
}

static void ad1816_reset (int dev)
{
  ad1816_halt (dev);
}


static int ad1816_set_speed (int dev, int arg)
{
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;

  if (arg == 0)
    return devc->speed;

  if (arg < 4000)
    arg = 4000;
  if (arg > 55000)
    arg = 55000;
  
  devc->speed = arg;
  return devc->speed;

}

static unsigned int ad1816_set_bits (int dev, unsigned int arg)
{
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;

  static struct format_tbl
    {
      int             format;
      unsigned char   bits;
    }
  format2bits[] =
  {
    {
      0, 0
    }
    ,
    {
      AFMT_MU_LAW, 1
    }
    ,
    {
      AFMT_A_LAW, 3
    }
    ,
    {
      AFMT_IMA_ADPCM, 0
    }
    ,
    {
      AFMT_U8, 0
    }
    ,
    {
      AFMT_S16_LE, 2
    }
    ,
    {
      AFMT_S16_BE, 6
    }
    ,
    {
      AFMT_S8, 0
    }
    ,
    {
      AFMT_U16_LE, 0
    }
    ,
    {
      AFMT_U16_BE, 0
    }
  };
  int             i, n = sizeof (format2bits) / sizeof (struct format_tbl);

  if (arg == 0)              /* return current format */
    return devc->audio_format;

  if (!(arg & ad_format_mask)) /* is this format supported? */
    arg = AFMT_U8;

  devc->audio_format = arg;

  for (i = 0; i < n; i++)
    if (format2bits[i].format == arg)
      {
	if ((devc->format_bits = format2bits[i].bits) == 0)
	  return devc->audio_format = AFMT_U8;	/* Was not supported */

	return arg;
      }

  /* Still hanging here. Something must be terribly wrong */
  devc->format_bits = 0;
  return devc->audio_format = AFMT_U8;
}

static short ad1816_set_channels (int dev, short arg)
{
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;

  if (arg != 1 && arg != 2)
    return devc->channels;

  devc->channels = arg;
  return arg;
}

static int ad1816_open (int dev, int mode) /* open device */
{
  ad1816_info    *devc = NULL;
  unsigned long   flags;

  if (dev < 0 || dev >= num_audiodevs) /* is device number valid ? */
    return -(ENXIO);

  devc = (ad1816_info *) audio_devs[dev]->devc; /* get device info of this dev */


  save_flags (flags); /* make check if device already open atomic */
  cli ();
  if (devc->opened)
    {
      restore_flags (flags);
      return -(EBUSY);
    }

  devc->opened = 1; /* mark device as open */

  devc->dual_dma = 0;

  if (audio_devs[dev]->flags & DMA_DUPLEX) /* if we are setup for duplex */
    {
      devc->dual_dma = 1;                  /* set dual_dma = 1 */
    }

  devc->intr_active = 0;                  /* default setup */
  devc->opened = 1;
  devc->audio_mode = 0;
  devc->speed = 8000;
  devc->audio_format=AFMT_U8;

  ad1816_reset(devc->dev_no); /* halt all pending output */
  restore_flags (flags);
  return 0;
}

static void ad1816_close (int dev) /* close device */
{
  unsigned long flags;
  ad1816_info    *devc = (ad1816_info *) audio_devs[dev]->devc;

  save_flags (flags); /* make register access atomic */
  cli ();
  ad1816_reset(devc->dev_no); /* halt all pending output */

  devc->intr_active = 0; /* default setup */
  devc->opened = 0;
  devc->audio_mode = 0;
  devc->speed = 8000;
  devc->audio_format=AFMT_U8;
  restore_flags (flags);
}


/* ------------------------------------------------------------------- */

/* Audio driver structure */

static struct audio_driver ad1816_audio_driver =
{
  ad1816_open,
  ad1816_close,
  ad1816_output_block,
  ad1816_start_input,
  ad1816_ioctl,
  ad1816_prepare_for_input,
  ad1816_prepare_for_output,
  ad1816_reset,
  ad1816_halt,
  NULL,
  NULL,
  ad1816_halt_input,
  ad1816_halt_output,
  ad1816_trigger,
  ad1816_set_speed,
  ad1816_set_bits,
  ad1816_set_channels
};


/* ------------------------------------------------------------------- */

/* Interrupt handler */

void ad1816_interrupt (int irq, void *dev_id, struct pt_regs *dummy)
{
  unsigned char   status;
  ad1816_info    *devc;
  int             dev;
  unsigned long flags;

  if (irq < 0 || irq > 15)
    {
      dev = -1;
    }
  else
    dev = irq2dev[irq];

  if (dev < 0 || dev >= num_audiodevs)
    {
      for (irq = 0; irq < 17; irq++)
	if (irq2dev[irq] != -1)
	  break;
      
      if (irq > 15)
	{
	  printk ("ad1848.c: Bogus interrupt %d\n", irq);
	  return;
	}

      dev = irq2dev[irq];
      devc = (ad1816_info *) audio_devs[dev]->devc;
    }
  else
    devc = (ad1816_info *) audio_devs[dev]->devc;

  save_flags(flags);
  cli();

  status = inb (devc->base+1); /* read interrupt register */
  outb (~status, devc->base+1);	/* Clear interrupt status */

  DEBUGINFO (printk("Got interrupt subclass %d\n",status));

  devc->irq_ok=1;

  if (status == 0)
    DEBUGWARN(printk ("ad1816_interrupt: Got interrupt, but no reason?\n"));
  
  if (devc->opened && devc->audio_mode & PCM_ENABLE_INPUT && status & 64)
    {
      DMAbuf_inputintr (dev);
    }

  if (devc->opened && devc->audio_mode & PCM_ENABLE_OUTPUT &&
      status & 128)
    {
      DMAbuf_outputintr (dev, 1);
    }
  restore_flags(flags);
}

/* ------------------------------------------------------------------- */

/* Mixer stuff */

struct mixer_def {
	unsigned int regno: 7;
	unsigned int polarity:1;	/* 0=normal, 1=reversed */
	unsigned int bitpos:4;
	unsigned int nbits:4;
};

static char mix_cvt[101] = {
	0, 0,3,7,10,13,16,19,21,23,26,28,30,32,34,35,37,39,40,42,
	43,45,46,47,49,50,51,52,53,55,56,57,58,59,60,61,62,63,64,65,
	65,66,67,68,69,70,70,71,72,73,73,74,75,75,76,77,77,78,79,79,
	80,81,81,82,82,83,84,84,85,85,86,86,87,87,88,88,89,89,90,90,
	91,91,92,92,93,93,94,94,95,95,96,96,96,97,97,98,98,98,99,99,
	100
};

typedef struct mixer_def mixer_ent;

/*
 * Most of the mixer entries work in backwards. Setting the polarity field
 * makes them to work correctly.
 *
 * The channel numbering used by individual soundcards is not fixed. Some
 * cards have assigned different meanings for the AUX1, AUX2 and LINE inputs.
 * The current version doesn't try to compensate this.
 */

#define MIX_ENT(name, reg_l, pola_l, pos_l, len_l, reg_r, pola_r, pos_r, len_r)	\
	{{reg_l, pola_l, pos_l, len_l}, {reg_r, pola_r, pos_r, len_r}}


mixer_ent mix_devices[32][2] = {
MIX_ENT(SOUND_MIXER_VOLUME,	14, 1, 8, 5,	14, 1, 0, 5),
MIX_ENT(SOUND_MIXER_BASS,	 0, 0, 0, 0,	 0, 0, 0, 0),
MIX_ENT(SOUND_MIXER_TREBLE,	 0, 0, 0, 0,	 0, 0, 0, 0),
MIX_ENT(SOUND_MIXER_SYNTH,	 5, 1, 8, 6,	 5, 1, 0, 6),
MIX_ENT(SOUND_MIXER_PCM,	 4, 1, 8, 6,	 4, 1, 0, 6),
MIX_ENT(SOUND_MIXER_SPEAKER,	 0, 0, 0, 0,	 0, 0, 0, 0),
MIX_ENT(SOUND_MIXER_LINE,	18, 1, 8, 5,	18, 1, 0, 5),
MIX_ENT(SOUND_MIXER_MIC,	19, 1, 8, 5,	19, 1, 0, 5),
MIX_ENT(SOUND_MIXER_CD,	 	15, 1, 8, 5,	15, 1, 0, 5),
MIX_ENT(SOUND_MIXER_IMIX,	 0, 0, 0, 0,	 0, 0, 0, 0),
MIX_ENT(SOUND_MIXER_ALTPCM,	 0, 0, 0, 0,	 0, 0, 0, 0),
MIX_ENT(SOUND_MIXER_RECLEV,	20, 0, 8, 4,	20, 0, 0, 4),
MIX_ENT(SOUND_MIXER_IGAIN,	 0, 0, 0, 0,	 0, 0, 0, 0),
MIX_ENT(SOUND_MIXER_OGAIN,	 0, 0, 0, 0,	 0, 0, 0, 0),
MIX_ENT(SOUND_MIXER_LINE1, 	17, 1, 8, 5,	17, 1, 0, 5),
MIX_ENT(SOUND_MIXER_LINE2,	16, 1, 8, 5,	16, 1, 0, 5),
MIX_ENT(SOUND_MIXER_LINE3,	 0, 0, 0, 0,	 0, 0, 0, 0)
};


static unsigned short default_mixer_levels[SOUND_MIXER_NRDEVICES] =
{
  0x6464,			/* Master Volume */
  0x3232,			/* Bass */
  0x3232,			/* Treble */
  0x0000,			/* FM */
  0x6464,			/* PCM */
  0x0000,			/* PC Speaker */
  0x0000,			/* Ext Line */
  0x0000,			/* Mic */
  0x0000,			/* CD */
  0x0000,			/* Recording monitor */
  0x0000,			/* SB PCM */
  0x0000,			/* Recording level */
  0x4b4b,			/* Input gain */
  0x4b4b,			/* Output gain */
  0x0000,			/* Line1 */
  0x0000,			/* Line2 */
  0x1515			/* Line3 (usually line in)*/
};

#define LEFT_CHN	0
#define RIGHT_CHN	1



static int
ad1816_set_recmask (ad1816_info * devc, int mask)
{
  unsigned char   recdev;
  int             i, n;

  mask &= devc->supported_rec_devices;

  n = 0;
  for (i = 0; i < 32; i++)	/* Count selected device bits */
    if (mask & (1 << i))
      n++;

  if (n == 0)
    mask = SOUND_MASK_MIC;
  else if (n != 1)		/* Too many devices selected */
    {
      mask &= ~devc->recmask;	/* Filter out active settings */

      n = 0;
      for (i = 0; i < 32; i++)	/* Count selected device bits */
	if (mask & (1 << i))
	  n++;

      if (n != 1)
	mask = SOUND_MASK_MIC;
    }

  switch (mask)
    {
    case SOUND_MASK_MIC:
      recdev = 5;
      break;

    case SOUND_MASK_LINE:
      recdev = 0;
      break;

    case SOUND_MASK_CD:
      recdev = 2;
      break;

    case SOUND_MASK_IMIX:
      recdev = 1;
      break;

    default:
      mask = SOUND_MASK_MIC;
      recdev = 5;
    }

  recdev <<= 4;
  ad_write (devc, 20, (ad_read (devc, 20) & 0x8f8f) | recdev | (recdev<<8));

  devc->recmask = mask;
  return mask;
}

static void
change_bits (int *regval, int dev, int chn, int newval)
{
  unsigned char   mask;
  int             shift;

  if (mix_devices[dev][chn].polarity == 1)	/* Reverse */
    newval = 100 - newval;

  mask = (1 << mix_devices[dev][chn].nbits) - 1;
  shift = mix_devices[dev][chn].bitpos;
  newval = (int) ((newval * mask) + 50) / 100;	/* Scale it */

  *regval &= ~(mask << shift);	/* Clear bits */
  *regval |= (newval & mask) << shift;	/* Set new value */
}

static int
ad1816_mixer_get (ad1816_info * devc, int dev)
{
  DEBUGINFO(printk("ad1816_mixer_get called!\n"));
  if (!((1 << dev) & devc->supported_devices))
    return -(EINVAL);

  return devc->levels[dev];
}

static int
ad1816_mixer_set (ad1816_info * devc, int dev, int value)
{
  int             left = value & 0x000000ff;
  int             right = (value & 0x0000ff00) >> 8;
  int             retvol;

  int             regoffs;
  int   val, valmute;

  DEBUGINFO(printk("ad1816_mixer_set called!\n"));

  if (left > 100)
    left = 100;
  if (right > 100)
    right = 100;

  if (mix_devices[dev][RIGHT_CHN].nbits == 0)	/* Mono control */
    right = left;

  retvol = left | (right << 8);

  /* Scale volumes */
  /*
  left = mix_cvt[left];
  right = mix_cvt[right];
  */
  /* Scale it again */

  left = mix_cvt[left];
  right = mix_cvt[right];


  if (dev > 31)
    return -(EINVAL);

  if (!(devc->supported_devices & (1 << dev)))
    return -(EINVAL);

  if (mix_devices[dev][LEFT_CHN].nbits == 0)
    return -(EINVAL);

  devc->levels[dev] = retvol;

  /*
     * Set the left channel
   */
  regoffs = mix_devices[dev][LEFT_CHN].regno;
  val = ad_read (devc, regoffs);
  change_bits (&val, dev, LEFT_CHN, left);

  valmute=val;

  if ( regoffs==4 || regoffs==5 || regoffs==14 || regoffs==15 ||
       regoffs==16 || regoffs==17 || regoffs==18 || regoffs==19)
    {
      if (left==0)
	valmute |= 0x8000;
      else
	valmute &= ~0x8000;
    }
    ad_write (devc, regoffs, valmute); /* mute */

  devc->saved_regs[regoffs] = val;
  /*
     * Set the right channel
   */
 
  if (mix_devices[dev][RIGHT_CHN].nbits == 0)
    return retvol;		/* Was just a mono channel */
  regoffs = mix_devices[dev][RIGHT_CHN].regno;
  val = ad_read (devc, regoffs);
  change_bits (&val, dev, RIGHT_CHN, right);

  valmute=val;
  if ( regoffs==4 || regoffs==5 || regoffs==14 || regoffs==15 ||
       regoffs==16 || regoffs==17 || regoffs==18 || regoffs==19)
    {
      if (right==0)
	valmute |= 0x80;
      else
	valmute &= ~0x80;
    }
    ad_write (devc, regoffs, valmute); /* mute */
  

  devc->saved_regs[regoffs] = val;
  return retvol;
}

#define MIXER_DEVICES ( SOUND_MASK_VOLUME | \
			SOUND_MASK_SYNTH | \
			SOUND_MASK_PCM | \
			SOUND_MASK_LINE | \
			SOUND_MASK_LINE1 | \
			SOUND_MASK_LINE2 | \
			SOUND_MASK_MIC | \
			SOUND_MASK_CD | \
			SOUND_MASK_RECLEV  \
			)
#define REC_DEVICES ( SOUND_MASK_SYNTH |\
		      SOUND_MASK_LINE |\
		      SOUND_MASK_MIC |\
		      SOUND_MASK_CD \
		      )
     
static void
ad1816_mixer_reset (ad1816_info * devc)
{
  int             i;

  devc->supported_devices = MIXER_DEVICES;
  
  devc->supported_rec_devices = REC_DEVICES;

  for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
    if (devc->supported_devices & (1 << i))
      ad1816_mixer_set (devc, i, default_mixer_levels[i]);
  ad1816_set_recmask (devc, SOUND_MASK_MIC);
}

static int
ad1816_mixer_ioctl (int dev, unsigned int cmd, caddr_t arg)
{
  ad1816_info    *devc = mixer_devs[dev]->devc;

  DEBUGINFO(printk("ad1816_mixer_ioctl called!\n"));
  
  if (((cmd >> 8) & 0xff) == 'M')
    {

      if (_IOC_DIR (cmd) & _IOC_WRITE)
	switch (cmd & 0xff)
	  {
	  case SOUND_MIXER_RECSRC:
	    return snd_ioctl_return ((int *) arg, ad1816_set_recmask (devc, get_user ((int *) arg)));
	    break;

	  default:
	    return snd_ioctl_return ((int *) arg, ad1816_mixer_set (devc, cmd & 0xff, get_user ((int *) arg)));
	  }
      else
	switch (cmd & 0xff)	/*
				 * Return parameters
				 */
	  {

	  case SOUND_MIXER_RECSRC:
	    return snd_ioctl_return ((int *) arg, devc->recmask);
	    break;

	  case SOUND_MIXER_DEVMASK:
	    return snd_ioctl_return ((int *) arg, devc->supported_devices);
	    break;

	  case SOUND_MIXER_STEREODEVS:
	    return snd_ioctl_return ((int *) arg, devc->supported_devices & ~(SOUND_MASK_SPEAKER | SOUND_MASK_IMIX));
	    break;

	  case SOUND_MIXER_RECMASK:
	    return snd_ioctl_return ((int *) arg, devc->supported_rec_devices);
	    break;

	  case SOUND_MIXER_CAPS:
	    return snd_ioctl_return ((int *) arg, SOUND_CAP_EXCL_INPUT);
	    break;

	  default:
	    return snd_ioctl_return ((int *) arg, ad1816_mixer_get (devc, cmd & 0xff));
	  }
    }
  else
    return -(EINVAL);
}

/* ------------------------------------------------------------------- */

/* Mixer structure */

static struct mixer_operations ad1816_mixer_operations =
{
  "AD1816",
  "AD1816 Mixer",
  ad1816_mixer_ioctl
};


/* ------------------------------------------------------------------- */

/* stuff for card recognition, init and unloading */


/* replace with probe routine */
int probe_ad1816 ( struct address_info *hw_config )
{
  ad1816_info    *devc = &dev_info[nr_ad1816_devs];
  int io_base=hw_config->io_base;
  int *osp=hw_config->osp;
  int tmp;

  if (check_region (io_base, 16))
    {
      printk ("cs4232.c: I/O port 0x%03x not free\n", io_base);
      return 0;
    }

  DEBUGLOG(printk ("ad1848_detect(%x)\n", io_base));
  
  if (nr_ad1816_devs >= MAX_AUDIO_DEV)
    {
      printk ("ad1848 detect error - step 0\n");
      return 0;
    }
  if (check_region (io_base, 16))
    {
      printk ("\n\nad1848.c: Port %x not free.\n\n", io_base);
      return 0;
    }
  
  devc->base = io_base;
  devc->irq_ok = 0;
  devc->irq = 0;
  devc->opened = 0;
  devc->osp = osp;
  devc->debug_flag = 0;


  {
    tmp=inb(devc->base);
    if ( (tmp&0x80)==0 || tmp==255 )
      {
	DEBUGLOG (printk ("Chip is not an AD1816 or chip is not active (Test 0)\n"));
	return(0);
      }
    
  }

  ad_write(devc,8,12345); /* writes to ireg 8 are copied to ireg 9 */
  if (ad_read(devc,9)!=12345)
    {
	DEBUGLOG (printk ("Chip is not an AD1816 (Test 1)\n"));
	return(0);
    }
  
  ad_write(devc,8,54321); /* writes to ireg 8 are copied to ireg 9 */
  if (ad_read(devc,9)!=54321)
    {
	DEBUGLOG (printk ("Chip is not an AD1816 (Test 2)\n"));
	return(0);
    }
  
  ad_write(devc,10,54321); /* writes to ireg 10 are copied to ireg 11 */
  if (ad_read(devc,11)!=54321)
    {
	DEBUGLOG (printk ("Chip is not an AD1816 (Test 3)\n"));
	return(0);
    }

  ad_write(devc,10,12345); /* writes to ireg 10 are copied to ireg 11 */
  if (ad_read(devc,11)!=12345)
    {
	DEBUGLOG (printk ("Chip is not an AD1816 (Test 4)\n"));
	return(0);
    }

  tmp=inb(devc->base+1);
  outb(0xff,devc->base+1); /* bit in base +1 cannot be set to 1 */
  if (inb(devc->base+1)!=tmp)
    {
	DEBUGLOG (printk ("Chip is not an AD1816 (Test 5)\n"));
	return(0);
    }

  
  DEBUGLOG (printk ("ad1816_detect() - Detected OK\n"));
  DEBUGLOG (printk ("AD1816 Version: %d\n",ad_read(devc,45)));

  return 1; /* change to 1 if detection works */
}

/* replace with attach */
void attach_ad1816 (struct address_info *hw_config)
{
  /*
     * NOTE! If irq < 0, there is another driver which has allocated the IRQ
     *   so that this driver doesn't need to allocate/deallocate it.
     *   The actually used IRQ is ABS(irq).
   */


  int             my_dev;
  char            dev_name[100];
  ad1816_info    *devc = &dev_info[nr_ad1816_devs];
  int             audio_flags = DMA_AUTOMODE; /* ??? */

  int irq=hw_config->irq;
  int dma_playback=hw_config->dma;
  int dma_capture=hw_config->dma2;
  int *osp=hw_config->osp;

  request_region (devc->base, 16, "AD1816 Sound");

  devc->irq = (irq > 0) ? irq : 0;
  devc->opened = 0;
  devc->timer_ticks = 0;
  devc->osp = osp;

  if (dma_capture == dma_playback || dma_capture == -1 || dma_playback == -1)
    {
      /* Single DMA mode */
      audio_flags &= ~DMA_DUPLEX;
    }
  else
    {
      /* Dual DMA mode */
      audio_flags |= DMA_DUPLEX;
    }
  


  ad_write(devc,1,0);     /* disable all interrupts */
  outb (0, devc->base+1);	/* Clear pending interrupts */

  sprintf (dev_name,"AD1816 audio driver");
  
  conf_printf2 (dev_name,
		devc->base, devc->irq, dma_playback, dma_capture);

  if ((my_dev = sound_install_audiodrv (AUDIO_DRIVER_VERSION,
					dev_name,
					&ad1816_audio_driver,
					sizeof (struct audio_driver),
					audio_flags,
					ad_format_mask,
					devc,
					dma_playback,
					dma_capture)) < 0)
    {
      return;
    }

  if (irq > 0)
    {
      /* check for irq */
      irq2dev[irq] = devc->dev_no = my_dev;
      if (snd_set_irq_handler (devc->irq, ad1816_interrupt,
			       "SoundPort",
			       devc->osp) < 0)
	{
	  printk ("ad1848: IRQ in use\n");
	}
    }
  
  nr_ad1816_devs++;

  if (sound_alloc_dma (dma_playback, "Sound System"))
    printk ("ad1848.c: Can't allocate DMA%d\n", dma_playback);

  if (dma_capture != dma_playback)
    if (sound_alloc_dma (dma_capture, "Sound System (capture)"))
      printk ("ad1848.c: Can't allocate DMA%d\n", dma_capture);
    

  
  ad_write(devc,32,0x80f0); /* sound system mode */
  ad_write(devc,33,0x03f8); /* enable all audiosources for dsp */
  ad_write(devc,4,0x0000);  /* default values for volumes */
  ad_write(devc,5,0x0000);
  ad_write(devc,15,0x8888);
  ad_write(devc,16,0x8888);
  ad_write(devc,17,0x8888);
  ad_write(devc,18,0x8888);
  ad_write(devc,19,0xc888); /* +20db mic active */
  ad_write(devc,14,0x0000);
  outb(0x10,devc->base+8); /* set dma mode */
  outb(0x10,devc->base+9);

  ad_write(devc,1,0xc000); /* enable capture + playback interrupt */
  
  ad1816_mixer_reset (devc); /* set mixer defaults */
  
  if (sound_install_mixer (MIXER_DRIVER_VERSION,
			   dev_name,
			   &ad1816_mixer_operations,
			   sizeof (struct mixer_operations),
			   devc) >= 0)
    {
      audio_devs[my_dev]->mixer_dev = num_mixers - 1;
      audio_devs[my_dev]->min_fragment = 0;
    }
}



void unload_ad1816 (struct address_info *hw_config)
{
  int             i, dev = 0;
  ad1816_info    *devc = NULL;

  int io_base=hw_config->io_base;
  int irq=hw_config->irq;

  for (i = 0; devc == NULL && i < nr_ad1816_devs; i++)
    if (dev_info[i].base == io_base)
      {
	devc = &dev_info[i];
	dev = devc->dev_no;
      }
  
  if (devc != NULL)
    {
      release_region (devc->base, 16);

      if (irq > 0)
	snd_release_irq (devc->irq);
      
      sound_free_dma (audio_devs[dev]->dmachan1);

      if (audio_devs[dev]->dmachan2 != audio_devs[dev]->dmachan1)
	sound_free_dma (audio_devs[dev]->dmachan2);
    }
  else
    printk ("ad1848: Can't find device to be unloaded. Base=%x\n",
	    io_base);
}

#endif
